--- plugins/gssapi.c.orig	2012-01-27 23:31:36.000000000 +0000
+++ plugins/gssapi.c	2013-10-09 23:39:33.420263116 +0000
@@ -202,7 +202,8 @@ sasl_gss_seterror_(const sasl_utils_t *u
     OM_uint32 msg_ctx;
     int ret;
     char *out = NULL;
-    size_t len, curlen = 0;
+    size_t len;
+    unsigned curlen = 0;
     const char prefix[] = "GSSAPI Error: ";
 
     if (!utils) return SASL_OK;
@@ -1212,7 +1213,7 @@ gssapi_server_mech_ssfreq(context_t *tex
 					(OM_uint32) oparams->maxoutbuf,
 					&max_input);
 
-	if(max_input > oparams->maxoutbuf) {
+	if (max_input > oparams->maxoutbuf) {
 	    /* Heimdal appears to get this wrong */
 	    oparams->maxoutbuf -= (max_input - oparams->maxoutbuf);
 	} else {
@@ -1463,6 +1464,116 @@ static int gssapi_client_mech_new(void *
     return SASL_OK;
 }
 
+static int gssapi_client_ssf_choose(context_t *text,
+				    sasl_client_params_t *params,
+				    sasl_out_params_t *oparams,
+				    char serverhas,
+				    unsigned maxoutbuf,
+				    char *omychoice)
+{
+    sasl_security_properties_t *secprops = &(params->props);
+    unsigned int external = params->external_ssf;
+    sasl_ssf_t need, allowed;
+    OM_uint32 maj_stat = 0, min_stat = 0;
+    char mychoice;
+
+    /* taken from kerberos.c */
+    if (secprops->min_ssf > (K5_MAX_SSF + external)) {
+	return SASL_TOOWEAK;
+    } else if (secprops->min_ssf > secprops->max_ssf) {
+	return SASL_BADPARAM;
+    }
+
+    /* need bits of layer -- sasl_ssf_t is unsigned so be careful */
+    if (secprops->max_ssf >= external) {
+	allowed = secprops->max_ssf - external;
+    } else {
+	allowed = 0;
+    }
+    if (secprops->min_ssf >= external) {
+	need = secprops->min_ssf - external;
+    } else {
+	/* good to go */
+	need = 0;
+    }
+
+    /* use the strongest layer available */
+    if ((text->qop & LAYER_CONFIDENTIALITY) &&
+	allowed >= K5_MAX_SSF &&
+	need <= K5_MAX_SSF &&
+	(serverhas & LAYER_CONFIDENTIALITY)) {
+	    
+	/* encryption */
+	oparams->encode = &gssapi_privacy_encode;
+	oparams->decode = &gssapi_decode;
+	/* FIX ME: Need to extract the proper value here */
+	oparams->mech_ssf = K5_MAX_SSF;
+	mychoice = LAYER_CONFIDENTIALITY;
+
+	if (serverhas & LAYER_INTEGRITY) {
+	    const char *ad_compat;
+
+	    params->utils->getopt(params->utils->getopt_context,
+				  "GSSAPI",
+				  "ad_compat",
+				  &ad_compat,
+				  NULL);
+	    /* should we send an AD compatible choice of security layers? */
+	    if (ad_compat &&
+		(ad_compat[0] == '1' || ad_compat[0] == 'y' ||
+		(ad_compat[0] == 'o' && ad_compat[1] == 'n') ||
+		ad_compat[0] == 't')) {
+		mychoice = LAYER_INTEGRITY|LAYER_CONFIDENTIALITY;
+	    }
+	}
+    } else if ((text->qop & LAYER_INTEGRITY) &&
+		allowed >= 1 &&
+		need <= 1 &&
+		(serverhas & LAYER_INTEGRITY)) {
+	/* integrity */
+	oparams->encode = &gssapi_integrity_encode;
+	oparams->decode = &gssapi_decode;
+	oparams->mech_ssf = 1;
+	mychoice = LAYER_INTEGRITY;
+    } else if ((text->qop & LAYER_NONE) &&
+	       need <= 0 && (serverhas & LAYER_NONE)) {
+	/* no layer */
+	oparams->encode = NULL;
+	oparams->decode = NULL;
+	oparams->mech_ssf = 0;
+	mychoice = LAYER_NONE;
+    } else {
+	SETERROR(text->utils,
+		 "No mutually compatible encoding layer between client and server");
+	return SASL_TOOWEAK;
+    }
+	
+    oparams->maxoutbuf = maxoutbuf;
+
+    if (oparams->mech_ssf) {
+	OM_uint32 max_input;
+
+	maj_stat = gss_wrap_size_limit(&min_stat,
+				       text->gss_ctx,
+				       1,
+				       GSS_C_QOP_DEFAULT,
+				       (OM_uint32) oparams->maxoutbuf,
+				       &max_input);
+
+	if (max_input > oparams->maxoutbuf) {
+	    /* Heimdal appears to get this wrong */
+	    oparams->maxoutbuf -= (max_input - oparams->maxoutbuf);
+	} else {
+	    /* This code is actually correct */
+	    oparams->maxoutbuf = max_input;
+	}
+    }
+    if (omychoice != NULL) {
+	*omychoice = mychoice;
+    }
+    return SASL_OK;
+}
+
 static int gssapi_client_mech_step(void *conn_context,
 				   sasl_client_params_t *params,
 				   const char *serverin,
@@ -1732,6 +1843,48 @@ static int gssapi_client_mech_step(void 
 		return SASL_OK;
 	    }
 
+	    if (text->mech_type == &gss_spnego_oid) {
+		unsigned maxoutbuf;
+		int saslrc;
+		char serverhas;
+
+		params->utils->log(NULL, SASL_LOG_DEBUG,
+				   "GS-SPNEGO SSF verification out_req_flags 0x%x", out_req_flags);
+
+		/* bit mask of server support */
+		if (out_req_flags & GSS_C_CONF_FLAG)
+		{
+		    serverhas = LAYER_CONFIDENTIALITY;
+		}
+		else if (out_req_flags & GSS_C_INTEG_FLAG)
+		{
+		    serverhas = LAYER_INTEGRITY;
+		}
+		else
+		{
+		    serverhas = LAYER_NONE;
+		}
+		
+#define SEND_MAX_WIRE_SIZE      0x00A00000
+		saslrc = gssapi_client_ssf_choose(text,
+						  params,
+						  oparams,
+						  serverhas,
+						  SEND_MAX_WIRE_SIZE,
+						  /*omychoice*/NULL);
+		if (saslrc == SASL_OK) {
+		    text->state = SASL_GSSAPI_STATE_AUTHENTICATED;
+		
+		    oparams->doneflag = 1;
+		
+		    /* used by layers */
+		    _plug_decode_init(&text->decode_context, text->utils,
+				      (params->props.maxbufsize > 0xFFFFFF) ? 0xFFFFFF :
+				      params->props.maxbufsize);
+		}
+		return saslrc;	
+	    }
+
 	    /* Switch to ssf negotiation */
 	    text->state = SASL_GSSAPI_STATE_SSFCAP;
 	}
@@ -1739,9 +1892,9 @@ static int gssapi_client_mech_step(void 
 	return SASL_CONTINUE;
 
     case SASL_GSSAPI_STATE_SSFCAP: {
-	sasl_security_properties_t *secprops = &(params->props);
-	unsigned int alen, external = params->external_ssf;
-	sasl_ssf_t need, allowed;
+	int saslrc;
+	unsigned int alen;
+	unsigned maxoutbuf;
 	char serverhas, mychoice;
 	
 	real_input_token.value = (void *) serverin;
@@ -1777,106 +1930,27 @@ static int gssapi_client_mech_step(void 
 	    return SASL_FAIL;
 	}
 
-	/* taken from kerberos.c */
-	if (secprops->min_ssf > (K5_MAX_SSF + external)) {
-	    return SASL_TOOWEAK;
-	} else if (secprops->min_ssf > secprops->max_ssf) {
-	    return SASL_BADPARAM;
-	}
-	
-	/* need bits of layer -- sasl_ssf_t is unsigned so be careful */
-	if (secprops->max_ssf >= external) {
-	    allowed = secprops->max_ssf - external;
-	} else {
-	    allowed = 0;
-	}
-	if (secprops->min_ssf >= external) {
-	    need = secprops->min_ssf - external;
-	} else {
-	    /* good to go */
-	    need = 0;
-	}
-
 	/* bit mask of server support */
 	serverhas = ((char *)output_token->value)[0];
-	
-	/* use the strongest layer available */
-	if ((text->qop & LAYER_CONFIDENTIALITY) &&
-	    allowed >= K5_MAX_SSF &&
-	    need <= K5_MAX_SSF &&
-	    (serverhas & LAYER_CONFIDENTIALITY)) {
-	    
-	    const char *ad_compat;
-
-	    /* encryption */
-	    oparams->encode = &gssapi_privacy_encode;
-	    oparams->decode = &gssapi_decode;
-	    /* FIX ME: Need to extract the proper value here */
-	    oparams->mech_ssf = K5_MAX_SSF;
-	    mychoice = LAYER_CONFIDENTIALITY;
-
-	    if (serverhas & LAYER_INTEGRITY) {
-		/* should we send an AD compatible choice of security layers? */
-		params->utils->getopt(params->utils->getopt_context,
-				      "GSSAPI",
-				      "ad_compat",
-				      &ad_compat,
-				      NULL);
-		if (ad_compat &&
-		    (ad_compat[0] == '1' || ad_compat[0] == 'y' ||
-		     (ad_compat[0] == 'o' && ad_compat[1] == 'n') ||
-		     ad_compat[0] == 't')) {
-		    mychoice = LAYER_INTEGRITY|LAYER_CONFIDENTIALITY;
-		}
-	    }
-	} else if ((text->qop & LAYER_INTEGRITY) &&
-		    allowed >= 1 &&
-		    need <= 1 &&
-		    (serverhas & LAYER_INTEGRITY)) {
-	    /* integrity */
-	    oparams->encode = &gssapi_integrity_encode;
-	    oparams->decode = &gssapi_decode;
-	    oparams->mech_ssf = 1;
-	    mychoice = LAYER_INTEGRITY;
-	} else if ((text->qop & LAYER_NONE) &&
-		   need <= 0 && (serverhas & LAYER_NONE)) {
-	    /* no layer */
-	    oparams->encode = NULL;
-	    oparams->decode = NULL;
-	    oparams->mech_ssf = 0;
-	    mychoice = LAYER_NONE;
-	} else {
-	    /* there's no appropriate layering for us! */
-	    sasl_gss_free_context_contents(text);
-	    return SASL_TOOWEAK;
-	}
-	
-        oparams->maxoutbuf =
+        maxoutbuf =
 	    (((unsigned char *) output_token->value)[1] << 16) |
             (((unsigned char *) output_token->value)[2] << 8) |
             (((unsigned char *) output_token->value)[3] << 0);
-
-	if (oparams->mech_ssf) {
-            maj_stat = gss_wrap_size_limit( &min_stat,
-                                            text->gss_ctx,
-                                            1,
-                                            GSS_C_QOP_DEFAULT,
-                                            (OM_uint32) oparams->maxoutbuf,
-                                            &max_input);
-
-	    if (max_input > oparams->maxoutbuf) {
-		/* Heimdal appears to get this wrong */
-		oparams->maxoutbuf -= (max_input - oparams->maxoutbuf);
-	    } else {
-		/* This code is actually correct */
-		oparams->maxoutbuf = max_input;
-	    }
-	}
-	
 	GSS_LOCK_MUTEX(params->utils);
 	gss_release_buffer(&min_stat, output_token);
 	GSS_UNLOCK_MUTEX(params->utils);
 	
+	saslrc = gssapi_client_ssf_choose(text,
+	    				  params,
+					  oparams,
+					  serverhas,
+					  maxoutbuf,
+					  &mychoice);
+	if (saslrc) {
+	    sasl_gss_free_context_contents(text);
+	    return saslrc;
+	}
+	
 	/* oparams->user is always set, due to canon_user requirements.
 	 * Make sure the client actually requested it though, by checking
 	 * if our context was set.
